#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>

#include <unistd.h>
#include <stdlib.h>

#include <sys/mman.h>

#define CR_ST_WORD(state, i)   (state.words[i])

union CRYPTO_STATE {
    uint8_t    bytes[16];
    uint32_t   words[4];
    uint64_t   l[2];
};

static uint8_t const sm4_sbox[] = {
    0xd6, 0x90, 0xe9, 0xfe, 0xcc, 0xe1, 0x3d, 0xb7,
    0x16, 0xb6, 0x14, 0xc2, 0x28, 0xfb, 0x2c, 0x05,
    0x2b, 0x67, 0x9a, 0x76, 0x2a, 0xbe, 0x04, 0xc3,
    0xaa, 0x44, 0x13, 0x26, 0x49, 0x86, 0x06, 0x99,
    0x9c, 0x42, 0x50, 0xf4, 0x91, 0xef, 0x98, 0x7a,
    0x33, 0x54, 0x0b, 0x43, 0xed, 0xcf, 0xac, 0x62,
    0xe4, 0xb3, 0x1c, 0xa9, 0xc9, 0x08, 0xe8, 0x95,
    0x80, 0xdf, 0x94, 0xfa, 0x75, 0x8f, 0x3f, 0xa6,
    0x47, 0x07, 0xa7, 0xfc, 0xf3, 0x73, 0x17, 0xba,
    0x83, 0x59, 0x3c, 0x19, 0xe6, 0x85, 0x4f, 0xa8,
    0x68, 0x6b, 0x81, 0xb2, 0x71, 0x64, 0xda, 0x8b,
    0xf8, 0xeb, 0x0f, 0x4b, 0x70, 0x56, 0x9d, 0x35,
    0x1e, 0x24, 0x0e, 0x5e, 0x63, 0x58, 0xd1, 0xa2,
    0x25, 0x22, 0x7c, 0x3b, 0x01, 0x21, 0x78, 0x87,
    0xd4, 0x00, 0x46, 0x57, 0x9f, 0xd3, 0x27, 0x52,
    0x4c, 0x36, 0x02, 0xe7, 0xa0, 0xc4, 0xc8, 0x9e,
    0xea, 0xbf, 0x8a, 0xd2, 0x40, 0xc7, 0x38, 0xb5,
    0xa3, 0xf7, 0xf2, 0xce, 0xf9, 0x61, 0x15, 0xa1,
    0xe0, 0xae, 0x5d, 0xa4, 0x9b, 0x34, 0x1a, 0x55,
    0xad, 0x93, 0x32, 0x30, 0xf5, 0x8c, 0xb1, 0xe3,
    0x1d, 0xf6, 0xe2, 0x2e, 0x82, 0x66, 0xca, 0x60,
    0xc0, 0x29, 0x23, 0xab, 0x0d, 0x53, 0x4e, 0x6f,
    0xd5, 0xdb, 0x37, 0x45, 0xde, 0xfd, 0x8e, 0x2f,
    0x03, 0xff, 0x6a, 0x72, 0x6d, 0x6c, 0x5b, 0x51,
    0x8d, 0x1b, 0xaf, 0x92, 0xbb, 0xdd, 0xbc, 0x7f,
    0x11, 0xd9, 0x5c, 0x41, 0x1f, 0x10, 0x5a, 0xd8,
    0x0a, 0xc1, 0x31, 0x88, 0xa5, 0xcd, 0x7b, 0xbd,
    0x2d, 0x74, 0xd0, 0x12, 0xb8, 0xe5, 0xb4, 0xb0,
    0x89, 0x69, 0x97, 0x4a, 0x0c, 0x96, 0x77, 0x7e,
    0x65, 0xb9, 0xf1, 0x09, 0xc5, 0x6e, 0xc6, 0x84,
    0x18, 0xf0, 0x7d, 0xec, 0x3a, 0xdc, 0x4d, 0x20,
    0x79, 0xee, 0x5f, 0x3e, 0xd7, 0xcb, 0x39, 0x48,
};

void myprint_chars(uint8_t *vd)
{
    for (int i=0; i<16; i++) {
        printf("%c", vd[i]);
    }
}

int32_t rol32(uint32_t v, unsigned int n)
{
    return (v << n) | (v >> (32 - n));
}

void crypto_sm4e(void *vd, void *vn)
{
    uint64_t *rd = vd;
    uint64_t *rn = vn;
    union CRYPTO_STATE d = { .l = { rd[0], rd[1] } };
    union CRYPTO_STATE n = { .l = { rn[0], rn[1] } };
    uint32_t t, i;

    for (i = 0; i < 4; i++) {
        t = CR_ST_WORD(d, (i + 1) % 4) ^
            CR_ST_WORD(d, (i + 2) % 4) ^
            CR_ST_WORD(d, (i + 3) % 4) ^
            CR_ST_WORD(n, i);

        t = sm4_sbox[t & 0xff] |
            sm4_sbox[(t >> 8) & 0xff] << 8 |
            sm4_sbox[(t >> 16) & 0xff] << 16 |
            sm4_sbox[(t >> 24) & 0xff] << 24;

        CR_ST_WORD(d, i) ^= t ^ rol32(t, 2) ^ rol32(t, 10) ^ rol32(t, 18) ^
                            rol32(t, 24);
    }

    rd[0] = d.l[0];
    rd[1] = d.l[1];
}

void crypto_sm4ekey(void *vd, void *vn, void* vm)
{
    uint64_t *rd = vd;
    uint64_t *rn = vn;
    uint64_t *rm = vm;
    union CRYPTO_STATE d;
    union CRYPTO_STATE n = { .l = { rn[0], rn[1] } };
    union CRYPTO_STATE m = { .l = { rm[0], rm[1] } };
    uint32_t t, i;

    d = n;
    for (i = 0; i < 4; i++) {
        t = CR_ST_WORD(d, (i + 1) % 4) ^
            CR_ST_WORD(d, (i + 2) % 4) ^
            CR_ST_WORD(d, (i + 3) % 4) ^
            CR_ST_WORD(m, i);

        t = sm4_sbox[t & 0xff] |
            sm4_sbox[(t >> 8) & 0xff] << 8 |
            sm4_sbox[(t >> 16) & 0xff] << 16 |
            sm4_sbox[(t >> 24) & 0xff] << 24;

        CR_ST_WORD(d, i) ^= t ^ rol32(t, 13) ^ rol32(t, 23);
    }

    rd[0] = d.l[0];
    rd[1] = d.l[1];
}

void rev32_word(uint32_t *w)
{
    uint8_t *u = (uint8_t *) w;
    *w = u[3]
        | (u[2] << 8)
        | (u[1] << 16)
        | (u[0] << 24);
}

void rev32(void *vn)
{
    uint64_t *rn = vn;
    union CRYPTO_STATE n = { .l = { rn[0], rn[1] } };

    for (int i=0; i<4; i++)
    {
        rev32_word(&n.words[i]);
    }

    rn[0] = n.l[0];
    rn[1] = n.l[1];
}

void decrypt(uint32_t offset, uint8_t *src, uint8_t bytes[16])
{
    uint64_t SM4_KEY[2] = {0xb4c150382078cc93, 0x625f824d5dc439c};
    uint64_t SM4_CK[2] = {0x1c232a3100070e15, 0x545b6269383f464d};

    uint64_t vd[2];
    memcpy(vd, src, 16);
    uint32_t v2_32[4] = {offset, offset, offset, offset};
    uint64_t *v2 = (uint64_t *) v2_32;
    SM4_CK[0] = SM4_CK[0] ^ v2[0];
    SM4_CK[1] = SM4_CK[1] ^ v2[1];
    crypto_sm4ekey(SM4_KEY, SM4_KEY, SM4_CK);

    union CRYPTO_STATE n = { .l = { SM4_KEY[0], SM4_KEY[1] } };
    union CRYPTO_STATE m;
    m.words[0] = n.words[3];
    m.words[1] = n.words[2];
    m.words[2] = n.words[1];
    m.words[3] = n.words[0];

    uint64_t *vn = m.l;

    rev32(vd);
    crypto_sm4e(vd, vn);
    rev32(vd);

    union CRYPTO_STATE p = { .l = { vd[0], vd[1] } };
    m.words[0] = p.words[3];
    m.words[1] = p.words[2];
    m.words[2] = p.words[1];
    m.words[3] = p.words[0];

    memcpy(bytes, m.bytes, 16);
}

void encrypt(uint32_t offset, uint8_t *src, uint8_t bytes[16])
{
    uint64_t SM4_KEY[2] = {0xb4c150382078cc93, 0x625f824d5dc439c};
    uint64_t SM4_CK[2] = {0x1c232a3100070e15, 0x545b6269383f464d};

    uint64_t vd[2];
    memcpy(vd, src, 16);
    uint32_t v2_32[4] = {offset, offset, offset, offset};
    uint64_t *v2 = (uint64_t *) v2_32;
    SM4_CK[0] = SM4_CK[0] ^ v2[0];
    SM4_CK[1] = SM4_CK[1] ^ v2[1];
    crypto_sm4ekey(SM4_KEY, SM4_KEY, SM4_CK);

    rev32(vd);
    crypto_sm4e(vd, SM4_KEY);
    rev32(vd);

    union CRYPTO_STATE n = { .l = { vd[0], vd[1] } };
    union CRYPTO_STATE m;
    m.words[0] = n.words[3];
    m.words[1] = n.words[2];
    m.words[2] = n.words[1];
    m.words[3] = n.words[0];
    memcpy(bytes, m.bytes, 16);
}

struct payload {
    uint8_t *base;
    size_t size;
};

static int file_load(const char *path, struct payload *payload)
{
    uint8_t *buf;
    struct stat stat;
    int fd;
    int ret;

    if (!path || !payload)
        return -EINVAL;

    fd = open(path, O_RDONLY|O_CLOEXEC);
    if (fd == -1) {
        ret = -errno;
        fprintf(stderr, "Failed to open '%s': %d(%m)\n", path, errno);
        return ret;
    }

    ret = fstat(fd, &stat);
    if (ret == -1) {
        ret = -errno;
        fprintf(stderr, "fstat() failed: %d(%m)\n", errno);
        goto close_fd;
    }

    buf = mmap(NULL, stat.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
    if (buf == MAP_FAILED) {
        ret = -errno;
        fprintf(stderr, "mmap() failed: %d(%m)\n", errno);
        goto close_fd;
    }

    payload->base = buf;
    payload->size = stat.st_size;

    ret = 0;

close_fd:
    close(fd);
    return ret;
}

int main(int argc, char **argv)
{
    int do_decrypt = 1;
    if (argc < 2) {
        fprintf(stderr, "missing -d/-e argument\n");
        exit(-1);
    }
    if (strcmp(argv[1], "-e") == 0) {
        do_decrypt = 0;
    } else if (strcmp(argv[1], "-d") != 0) {
        fprintf(stderr, "missing -d/-e argument\n");
        exit(-1);
    }
    struct payload payload;
    int fd = file_load(argv[2], &payload);

    uint8_t bytes[16];
    for (uint32_t offset=0; offset < payload.size; offset+=16) {
        memset(bytes, 0, 16);
        if (do_decrypt) {
            decrypt(offset, &payload.base[offset], bytes);
        } else {
            encrypt(offset, &payload.base[offset], bytes);
        }
        myprint_chars(bytes);
    }
}
